# Copyright © 2022 Arrikto Inc.  All Rights Reserved.

"""Kale SDK.

This script creates a TF Keras distributed KFP step to solve MNIST.
"""

from kale.sdk import pipeline, step
from kale.distributed import tf as kale_tf


@step(name="load_data")
def load_data():
    """Create the Training dataset."""
    def _data_fn():
        import tensorflow as tf
        import tensorflow_datasets as tfds

        buffer_size = 10000
        batch_size = 128

        def scale(image, label):
            """Scale MNIST data from (0, 255] to (0., 1.]."""
            image = tf.cast(image, tf.float32)
            image /= 255
            return image, label

        datasets, _ = tfds.load(name='mnist', with_info=True,
                                as_supervised=True)

        train_dataset = (datasets['train']
                         .map(scale)
                         .cache()
                         .shuffle(buffer_size)
                         .repeat()
                         .batch(batch_size))

        return train_dataset

    return _data_fn


@step(name="create_model")
def create_model():
    """Define the model."""
    def _create_model():
        import tensorflow as tf

        model = tf.keras.models.Sequential()
        model.add(
            tf.keras.layers.Conv2D(32, (3, 3), activation='relu',
                                   input_shape=(28, 28, 1)))
        model.add(tf.keras.layers.MaxPooling2D((2, 2)))
        model.add(tf.keras.layers.Conv2D(64, (3, 3), activation='relu'))
        model.add(tf.keras.layers.MaxPooling2D((2, 2)))
        model.add(tf.keras.layers.Conv2D(64, (3, 3), activation='relu'))
        model.add(tf.keras.layers.Flatten())
        model.add(tf.keras.layers.Dense(64, activation='relu'))
        model.add(tf.keras.layers.Dense(10, activation='softmax'))

        model.summary()

        model.compile(optimizer='adam',
                      loss='sparse_categorical_crossentropy',
                      metrics=['accuracy'])

        return model

    return _create_model


@step(name="model_training")
def distribute_training(model_fn, data_fn):
    """Train the model."""
    tfjob = kale_tf.distribute(
        model_fn=model_fn,
        train_data_fn=data_fn,
        epochs=4,
        number_of_processes=2,
        cuda=True,
        fit_kwargs={"steps_per_epoch": 70}
    )

    return tfjob.name


@pipeline(name="fmnist", experiment="kale-dist")
def ml_pipeline():
    """Run the ML pipeline."""
    data_fn = load_data()
    model_fn = create_model()
    distribute_training(model_fn, data_fn)


if __name__ == "__main__":
    ml_pipeline()
